@@ -1,3 +1,7 @@
+#ifdef __gnu_linux__
+#  include <sys/prctl.h>
+#endif
+
 #include "Python.h"
 #include <frameobject.h>
 
@@ -7,11 +11,36 @@
 #  define PYSTRING_CHECK PyString_Check
 #endif
 
-#define PUTS(fd, str) write(fd, str, (int)strlen(str))
+#define PUTS(fd, str) _Py_write_noraise(fd, str, (int)strlen(str))
 #define MAX_STRING_LENGTH 500
 #define MAX_FRAME_DEPTH 100
 #define MAX_NTHREADS 100
 
+/* Write count bytes of buf into fd.
+ *
+ * On success, return the number of written bytes, it can be lower than count
+ * including 0. On error, set errno and return -1.
+ *
+ * When interrupted by a signal (write() fails with EINTR), retry the syscall
+ * without calling the Python signal handler. */
+Py_ssize_t
+_Py_write_noraise(int fd, const char *buf, size_t count)
+{
+    Py_ssize_t res;
+
+    do {
+#ifdef MS_WINDOWS
+        assert(count < INT_MAX);
+        res = write(fd, buf, (int)count);
+#else
+        res = write(fd, buf, count);
+#endif
+    /* retry write() if it was interrupted by a signal */
+    } while (res < 0 && errno == EINTR);
+
+    return res;
+}
+
 /* Reverse a string. For example, "abcd" becomes "dcba".
 
    This function is signal safe. */
@@ -49,7 +78,7 @@ dump_decimal(int fd, int value)
         len++;
     } while (value);
     reverse_string(buffer, len);
-    write(fd, buffer, len);
+    _Py_write_noraise(fd, buffer, len);
 }
 
 /* Format an integer in range [0; 0xffffffff] to hexadecimal of 'width' digits,
@@ -57,11 +86,12 @@ dump_decimal(int fd, int value)
 
    This function is signal safe. */
 
-static void
-dump_hexadecimal(int fd, unsigned long value, int width)
+void
+_Py_dump_hexadecimal(int fd, unsigned long value, size_t bytes)
 {
     const char *hexdigits = "0123456789abcdef";
-    int len;
+    size_t width = bytes * 2;
+    size_t len;
     char buffer[sizeof(unsigned long) * 2 + 1];
     len = 0;
     do {
@@ -70,7 +100,7 @@ dump_hexadecimal(int fd, unsigned long value, int width)
         len++;
     } while (len < width || value);
     reverse_string(buffer, len);
-    write(fd, buffer, len);
+    _Py_write_noraise(fd, buffer, len);
 }
 
 /* Write an unicode object into the file fd using ascii+backslashreplace.
@@ -108,11 +138,11 @@ dump_ascii(int fd, PyObject *text)
         if (' ' <= ch && ch < 0x7f) {
             /* printable ASCII character */
             char c = (char)ch;
-            write(fd, &c, 1);
+            _Py_write_noraise(fd, &c, 1);
         }
         else if (ch <= 0xff) {
             PUTS(fd, "\\x");
-            dump_hexadecimal(fd, ch, 2);
+            _Py_dump_hexadecimal(fd, ch, 1);
         }
         else
 #ifdef Py_UNICODE_WIDE
@@ -120,12 +150,12 @@ dump_ascii(int fd, PyObject *text)
 #endif
         {
             PUTS(fd, "\\u");
-            dump_hexadecimal(fd, ch, 4);
+            _Py_dump_hexadecimal(fd, ch, 2);
 #ifdef Py_UNICODE_WIDE
         }
         else {
             PUTS(fd, "\\U");
-            dump_hexadecimal(fd, ch, 8);
+            _Py_dump_hexadecimal(fd, ch, 4);
 #endif
         }
     }
@@ -134,11 +164,11 @@ dump_ascii(int fd, PyObject *text)
         ch = *s;
         if (' ' <= ch && ch <= 126) {
             /* printable ASCII character */
-            write(fd, s, 1);
+            _Py_write_noraise(fd, s, 1);
         }
         else {
             PUTS(fd, "\\x");
-            dump_hexadecimal(fd, ch, 2);
+            _Py_dump_hexadecimal(fd, ch, 1);
         }
     }
 #endif
@@ -161,9 +191,9 @@ dump_frame(int fd, PyFrameObject *frame)
     if (code != NULL && code->co_filename != NULL
         && PYSTRING_CHECK(code->co_filename))
     {
-        write(fd, "\"", 1);
+        PUTS(fd, "\"");
         dump_ascii(fd, code->co_filename);
-        write(fd, "\"", 1);
+        PUTS(fd, "\"");
     } else {
         PUTS(fd, "???");
     }
@@ -185,7 +215,7 @@ dump_frame(int fd, PyFrameObject *frame)
     else
         PUTS(fd, "???");
 
-    write(fd, "\n", 1);
+    PUTS(fd, "\n");
 }
 
 static void
@@ -215,19 +245,12 @@ dump_traceback(int fd, PyThreadState *tstate, int write_header)
     }
 }
 
-/* Write the current Python traceback into the file 'fd':
-
-   Traceback (most recent call first):
-     File "xxx", line xxx in <xxx>
-     File "xxx", line xxx in <xxx>
-     ...
-     File "xxx", line xxx in <xxx>
-
-   Write only the first MAX_FRAME_DEPTH frames. If the traceback is truncated, write
-   the line "  ...".
-
-   This function is signal safe. */
+/* Dump the traceback of a Python thread into fd. Use write() to write the
+   traceback and retry if write() is interrupted by a signal (failed with
+   EINTR), but don't call the Python signal handler.
 
+   The caller is responsible to call PyErr_CheckSignals() to call Python signal
+   handlers if signals were received. */
 void
 _Py_DumpTraceback(int fd, PyThreadState *tstate)
 {
@@ -246,18 +269,31 @@ write_thread_id(int fd, PyThreadState *tstate, int is_current)
         PUTS(fd, "Current thread 0x");
     else
         PUTS(fd, "Thread 0x");
-    dump_hexadecimal(fd, (unsigned long)tstate->thread_id, sizeof(unsigned long)*2);
+    _Py_dump_hexadecimal(fd, (unsigned long)tstate->thread_id, sizeof(unsigned long));
+
+#ifdef __gnu_linux__
+    /* Linux only, get and print thread name */
+    static char thread_name[16];
+    if (0 == prctl(PR_GET_NAME, (unsigned long) thread_name, 0, 0, 0)) {
+        if (0 != strlen(thread_name)) {
+            PUTS(fd, " <");
+            PUTS(fd, thread_name);
+            PUTS(fd, ">");
+        }
+    }
+#endif
+
     PUTS(fd, " (most recent call first):\n");
 }
 
-/* Dump the traceback of all threads. Return NULL on success, or an error
-   message on error.
-
-   This function is signal safe. */
+/* Dump the traceback of all Python threads into fd. Use write() to write the
+   traceback and retry if write() is interrupted by a signal (failed with
+   EINTR), but don't call the Python signal handler.
 
+   The caller is responsible to call PyErr_CheckSignals() to call Python signal
+   handlers if signals were received. */
 const char*
-_Py_DumpTracebackThreads(int fd,
-                         PyInterpreterState *interp,
+_Py_DumpTracebackThreads(int fd, PyInterpreterState *interp,
                          PyThreadState *current_thread)
 {
     PyThreadState *tstate;
@@ -274,7 +310,7 @@ _Py_DumpTracebackThreads(int fd,
     do
     {
         if (nthreads != 0)
-            write(fd, "\n", 1);
+            PUTS(fd, "\n");
         if (nthreads >= MAX_NTHREADS) {
             PUTS(fd, "...\n");
             break;
